from __future__ import annotations

from collections.abc import Callable
from random import Random
from typing import TYPE_CHECKING, assert_never, overload
from uuid import UUID, uuid4

from utilities.random import get_state
from utilities.sentinel import Sentinel

if TYPE_CHECKING:
    from utilities.types import MaybeCallableUUIDLike, Seed


UUID_PATTERN = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
UUID_EXACT_PATTERN = f"^{UUID_PATTERN}$"


##


def get_uuid(seed: Seed | None = None, /) -> UUID:
    """Generate a UUID, possibly with a seed."""
    if seed is None:
        return uuid4()
    state = get_state(seed)
    return UUID(int=state.getrandbits(128), version=4)


##


@overload
def to_uuid(uuid: Sentinel, /, *, seed: Seed | None = None) -> Sentinel: ...
@overload
def to_uuid(
    uuid: MaybeCallableUUIDLike | None = get_uuid, /, *, seed: Seed | None = None
) -> UUID: ...
def to_uuid(
    uuid: MaybeCallableUUIDLike | None | Sentinel = get_uuid,
    /,
    *,
    seed: Seed | None = None,
) -> UUID | Sentinel:
    """Convert to a UUID."""
    match uuid:
        case UUID() | Sentinel():
            return uuid
        case None:
            return get_uuid(seed)
        case str():
            return UUID(uuid)
        case int() | float() | bytes() | bytearray() | Random() as seed:
            return get_uuid(seed)
        case Callable() as func:
            return to_uuid(func(), seed=seed)
        case never:
            assert_never(never)


__all__ = ["UUID_EXACT_PATTERN", "UUID_PATTERN", "get_uuid", "to_uuid"]



from __future__ import annotations

import enum
import pathlib
from dataclasses import dataclass
from enum import auto
from operator import attrgetter
from re import search
from typing import TYPE_CHECKING, Any, ClassVar

import whenever
from click import ParamType, argument, command, echo, option
from click.testing import CliRunner
from hypothesis import given
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    data,
    frozensets,
    integers,
    ip_addresses,
    lists,
    sampled_from,
    uuids,
)
from pytest import mark, param

from utilities.click import (
    CONTEXT_SETTINGS_HELP_OPTION_NAMES,
    UUID,
    Date,
    DateDelta,
    DateTimeDelta,
    Enum,
    FrozenSetChoices,
    FrozenSetEnums,
    FrozenSetInts,
    FrozenSetStrs,
    IPv4Address,
    IPv6Address,
    ListChoices,
    ListEnums,
    ListInts,
    ListStrs,
    MonthDay,
    Path,
    PlainDateTime,
    Time,
    TimeDelta,
    YearMonth,
    ZonedDateTime,
)
from utilities.hypothesis import (
    date_deltas,
    date_time_deltas,
    dates,
    month_days,
    paths,
    plain_date_times,
    text_ascii,
    time_deltas,
    times,
    year_months,
    zoned_date_times,
)
from utilities.text import join_strs, strip_and_dedent

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable


class TestContextSettingsHelpOptionNames:
    @given(help_=sampled_from(["-h", "--help"]))
    def test_main(self, *, help_: str) -> None:
        @command(**CONTEXT_SETTINGS_HELP_OPTION_NAMES)
        def cli() -> None: ...

        result = CliRunner().invoke(cli, [help_])
        assert result.exit_code == 0


class TestPath:
    def test_path(self, *, tmp_path: pathlib.Path) -> None:
        path_use = pathlib.Path("~", tmp_path)

        @command()
        @argument("path", type=Path())
        def cli(*, path: pathlib.Path) -> None:
            assert isinstance(path, pathlib.Path)
            assert path == path.expanduser()

        result = CliRunner().invoke(cli, [str(path_use)])
        assert result.exit_code == 0


class _ExampleEnum(enum.Enum):
    a = auto()
    b = auto()
    c = auto()


def _lift_serializer[T](
    serializer: Callable[[T], str], /, *, sort: bool = False
) -> Callable[[Iterable[T]], str]:
    def wrapped(values: Iterable[T], /) -> str:
        return join_strs(map(serializer, values), sort=sort)

    return wrapped


@dataclass(kw_only=True, slots=True)
class _Case[T]:
    param: ParamType
    name: str
    repr: str | None = None
    strategy: SearchStrategy[T]
    serialize: Callable[[T], str]
    failable: bool = False


class TestParameters:
    cases: ClassVar[list[_Case]] = [
        _Case(
            param=Date(),
            name="date",
            strategy=dates(),
            serialize=whenever.Date.format_common_iso,
            failable=True,
        ),
        _Case(
            param=DateDelta(),
            name="date delta",
            strategy=date_deltas(parsable=True),
            serialize=whenever.DateDelta.format_common_iso,
            failable=True,
        ),
        _Case(
            param=DateTimeDelta(),
            name="date-time delta",
            strategy=date_time_deltas(parsable=True),
            serialize=whenever.DateTimeDelta.format_common_iso,
            failable=True,
        ),
        _Case(
            param=Enum(_ExampleEnum),
            name="enum[_ExampleEnum]",
            repr="ENUM[_ExampleEnum]",
            strategy=sampled_from(_ExampleEnum),
            serialize=attrgetter("name"),
            failable=True,
        ),
        _Case(
            param=FrozenSetInts(),
            name="frozenset[integer]",
            repr="FROZENSET[INT]",
            strategy=frozensets(integers()),
            serialize=_lift_serializer(str, sort=True),
            failable=True,
        ),
        _Case(
            param=FrozenSetChoices(["a", "b", "c"]),
            name="frozenset[choice]",
            repr="FROZENSET[Choice(['a', 'b', 'c'])]",
            strategy=frozensets(sampled_from(["a", "b", "c"])),
            serialize=_lift_serializer(str, sort=True),
            failable=True,
        ),
        _Case(
            param=FrozenSetEnums(_ExampleEnum),
            name="frozenset[enum[_ExampleEnum]]",
            repr="FROZENSET[ENUM[_ExampleEnum]]",
            strategy=frozensets(sampled_from(_ExampleEnum)),
            serialize=_lift_serializer(attrgetter("name"), sort=True),
            failable=True,
        ),
        _Case(
            param=FrozenSetStrs(),
            name="frozenset[text]",
            repr="FROZENSET[STRING]",
            strategy=frozensets(text_ascii()),
            serialize=_lift_serializer(str, sort=True),
        ),
        _Case(
            param=IPv4Address(),
            name="ipv4 address",
            strategy=ip_addresses(v=4),
            serialize=str,
            failable=True,
        ),
        _Case(
            param=IPv6Address(),
            name="ipv6 address",
            strategy=ip_addresses(v=6),
            serialize=str,
            failable=True,
        ),
        _Case(
            param=ListChoices(["a", "b", "c"]),
            name="list[choice]",
            repr="LIST[Choice(['a', 'b', 'c'])]",
            strategy=lists(sampled_from(["a", "b", "c"])),
            serialize=_lift_serializer(str),
            failable=True,
        ),
        _Case(
            param=ListInts(),
            name="list[integer]",
            repr="LIST[INT]",
            strategy=lists(integers()),
            serialize=_lift_serializer(str),
            failable=True,
        ),
        _Case(
            param=ListEnums(_ExampleEnum),
            name="list[enum[_ExampleEnum]]",
            repr="LIST[ENUM[_ExampleEnum]]",
            strategy=lists(sampled_from(_ExampleEnum)),
            serialize=_lift_serializer(attrgetter("name")),
            failable=True,
        ),
        _Case(
            param=MonthDay(),
            name="month-day",
            strategy=month_days(),
            serialize=whenever.MonthDay.format_common_iso,
            failable=True,
        ),
        _Case(
            param=Path(), name="path", strategy=paths(), serialize=str, failable=False
        ),
        _Case(
            param=PlainDateTime(),
            name="plain date-time",
            strategy=plain_date_times(),
            serialize=whenever.PlainDateTime.format_common_iso,
            failable=True,
        ),
        _Case(
            param=Time(),
            name="time",
            strategy=times(),
            serialize=whenever.Time.format_common_iso,
            failable=True,
        ),
        _Case(
            param=TimeDelta(),
            name="time-delta",
            strategy=time_deltas(),
            serialize=whenever.TimeDelta.format_common_iso,
            failable=True,
        ),
        _Case(
            param=UUID(), name="uuid", strategy=uuids(), serialize=str, failable=True
        ),
        _Case(
            param=YearMonth(),
            name="year-month",
            strategy=year_months(),
            serialize=whenever.YearMonth.format_common_iso,
            failable=True,
        ),
        _Case(
            param=ZonedDateTime(),
            name="zoned date-time",
            strategy=zoned_date_times(),
            serialize=whenever.ZonedDateTime.format_common_iso,
            failable=True,
        ),
    ]

    @given(data=data())
    @mark.parametrize(
        ("param", "strategy", "serialize"),
        [param(c.param, c.strategy, c.serialize) for c in cases],
    )
    def test_default(
        self,
        *,
        data: DataObject,
        param: ParamType,
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], str],
    ) -> None:
        default = data.draw(strategy)

        @command()
        @option("--value", type=param, default=default)
        def cli(*, value: Any) -> None:
            echo(f"value = {serialize(value)}")

        result = CliRunner().invoke(cli, args=[])
        assert result.exit_code == 0
        assert result.stdout == f"value = {serialize(default)}\n"

    @given(data=data())
    @mark.parametrize(
        ("param", "strategy", "serialize"),
        [param(c.param, c.strategy, c.serialize) for c in cases],
    )
    def test_cli_value(
        self,
        *,
        data: DataObject,
        param: ParamType,
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], str],
    ) -> None:
        @command()
        @option("--value", type=param)
        def cli(*, value: Any) -> None:
            echo(f"value = {serialize(value)}")

        value = data.draw(strategy)
        ser = serialize(value)
        result = CliRunner().invoke(cli, args=[f"--value={ser}"])
        assert result.exit_code == 0
        assert result.stdout == f"value = {ser}\n"

    @mark.parametrize(
        ("param", "serialize"),
        [param(c.param, c.serialize) for c in cases if c.failable],
    )
    def test_cli_fail(
        self, *, param: ParamType, serialize: Callable[[Any], str]
    ) -> None:
        @command()
        @option("--value", type=param)
        def cli(*, value: Any) -> None:
            echo(f"value = {serialize(value)}")

        result = CliRunner().invoke(cli, args=["--value=invalid"])
        assert result.exit_code == 2

    @mark.parametrize(("param", "name"), [param(c.param, c.name) for c in cases])
    def test_name(self, *, param: ParamType, name: str) -> None:
        assert param.name == name

    @mark.parametrize(
        ("param", "repr_", "name"), [param(c.param, c.repr, c.name) for c in cases]
    )
    def test_repr(self, *, param: ParamType, repr_: str | None, name: str) -> None:
        expected = name.upper() if repr_ is None else repr_
        assert repr(param) == expected

    @mark.parametrize(
        "param",
        [param(ListEnums(_ExampleEnum)), param(FrozenSetEnums(_ExampleEnum))],
        ids=str,
    )
    def test_error_list_and_frozensets_parse(self, *, param: ParamType) -> None:
        @command()
        @option("--value", type=param, default=0)
        def cli(*, value: list[_ExampleEnum] | frozenset[_ExampleEnum]) -> None:
            echo(f"value = {value}")

        result = CliRunner().invoke(cli)
        assert result.exit_code == 2
        assert search(
            "Invalid value for '--value': Object '0' of type 'int' must be a string",
            result.stderr,
        )


class TestCLIHelp:
    @mark.parametrize(
        ("param", "expected"),
        [
            param(
                str,
                """
    Usage: cli [OPTIONS]

    Options:
      --value TEXT
      --help        Show this message and exit.
""",
            ),
            param(
                Enum(_ExampleEnum),
                """
    Usage: cli [OPTIONS]

    Options:
      --value [a,b,c]
      --help           Show this message and exit.
""",
            ),
            param(
                FrozenSetEnums(_ExampleEnum),
                """
    Usage: cli [OPTIONS]

    Options:
      --value [FROZENSET[a,b,c] SEP=,]
      --help                          Show this message and exit.
""",
            ),
            param(
                FrozenSetStrs(),
                """
    Usage: cli [OPTIONS]

    Options:
      --value [FROZENSET[TEXT] SEP=,]
      --help                          Show this message and exit.
""",
            ),
            param(
                ListEnums(_ExampleEnum),
                """
    Usage: cli [OPTIONS]

    Options:
      --value [LIST[a,b,c] SEP=,]
      --help                       Show this message and exit.
""",
            ),
            param(
                ListStrs(),
                """
    Usage: cli [OPTIONS]

    Options:
      --value [LIST[TEXT] SEP=,]
      --help                      Show this message and exit.
""",
            ),
        ],
    )
    def test_main(self, *, param: Any, expected: str) -> None:
        @command()
        @option("--value", type=param)
        def cli(*, value: Any) -> None:
            echo(f"value = {value}")

        result = CliRunner().invoke(cli, ["--help"])
        assert result.exit_code == 0
        expected = strip_and_dedent(expected, trailing=True)
        assert result.stdout == expected



from __future__ import annotations

from asyncio import sleep
from collections.abc import Callable
from functools import wraps
from io import StringIO
from logging import StreamHandler, getLogger
from re import search
from typing import TYPE_CHECKING, Literal

from eventkit import Event
from hypothesis import given
from hypothesis.strategies import sampled_from
from pytest import raises

from utilities.eventkit import (
    LiftedEvent,
    LiftListenerError,
    TypedEvent,
    add_listener,
    lift_listener,
)
from utilities.hypothesis import temp_paths, text_ascii

if TYPE_CHECKING:
    from pathlib import Path


class TestAddListener:
    @given(sync_or_async=sampled_from(["sync", "async"]))
    async def test_main(self, *, sync_or_async: Literal["sync", "async"]) -> None:
        event = Event()
        called = False
        match sync_or_async:
            case "sync":

                def listener_sync() -> None:
                    nonlocal called
                    called |= True

                _ = add_listener(event, listener_sync)
            case "async":

                async def listener_async() -> None:
                    nonlocal called
                    called |= True
                    await sleep(0.01)

                _ = add_listener(event, listener_async)

        event.emit()
        await sleep(0.01)
        assert called

    @given(root=temp_paths(), sync_or_async=sampled_from(["sync", "async"]))
    async def test_no_error_handler_but_run_into_error(
        self, *, root: Path, sync_or_async: Literal["sync", "async"]
    ) -> None:
        logger = getLogger(str(root))
        logger.addHandler(StreamHandler(buffer := StringIO()))
        event = Event()

        match sync_or_async:
            case "sync":

                def listener_sync() -> None: ...

                _ = add_listener(
                    event, listener_sync, logger=str(root), logger_allow_pytest=True
                )
            case "async":

                async def listener_async() -> None:
                    await sleep(0.01)

                _ = add_listener(
                    event, listener_async, logger=str(root), logger_allow_pytest=True
                )

        event.emit(None)
        await sleep(0.01)
        pattern = r"listener_a?sync\(\) takes 0 positional arguments but 1 was given"
        contents = buffer.getvalue()
        assert search(pattern, contents)

    @given(
        name=text_ascii(min_size=1), case=sampled_from(["sync", "async/sync", "async"])
    )
    async def test_with_error_handler(
        self, *, name: str, case: Literal["sync", "async/sync", "async"]
    ) -> None:
        event = Event(name=name)
        assert event.name() == name
        called = False
        log: set[tuple[str, type[BaseException]]] = set()

        def listener_sync(is_success: bool, /) -> None:  # noqa: FBT001
            if is_success:
                nonlocal called
                called |= True
            else:
                raise ValueError

        def error_sync(event: Event, exception: BaseException, /) -> None:
            nonlocal log
            log.add((event.name(), type(exception)))

        async def listener_async(is_success: bool, /) -> None:  # noqa: FBT001
            if is_success:
                nonlocal called
                called |= True
                await sleep(0.01)
            else:
                raise ValueError

        async def error_async(event: Event, exception: BaseException, /) -> None:
            nonlocal log
            log.add((event.name(), type(exception)))
            await sleep(0.01)

        match case:
            case "sync":
                _ = add_listener(event, listener_sync, error=error_sync)
            case "async/sync":
                _ = add_listener(event, listener_async, error=error_sync)
            case "async":
                _ = add_listener(event, listener_async, error=error_async)
        event.emit(True)  # noqa: FBT003
        await sleep(0.01)
        assert called
        assert log == set()
        event.emit(False)  # noqa: FBT003
        await sleep(0.01)
        assert log == {(name, ValueError)}

    @given(
        case=sampled_from([
            "no/sync",
            "no/async",
            "have/sync",
            "have/async/sync",
            "have/async",
        ])
    )
    async def test_ignore(
        self,
        *,
        case: Literal[
            "no/sync", "no/async", "have/sync", "have/async/sync", "have/async"
        ],
    ) -> None:
        event = Event()
        called = False
        log: set[tuple[str, type[BaseException]]] = set()

        def listener_sync(is_success: bool, /) -> None:  # noqa: FBT001
            if is_success:
                nonlocal called
                called |= True
            else:
                raise ValueError

        def error_sync(event: Event, exception: BaseException, /) -> None:
            nonlocal log
            log.add((event.name(), type(exception)))

        async def listener_async(is_success: bool, /) -> None:  # noqa: FBT001
            if is_success:
                nonlocal called
                called |= True
                await sleep(0.01)
            else:
                raise ValueError

        async def error_async(event: Event, exception: BaseException, /) -> None:
            nonlocal log
            log.add((event.name(), type(exception)))
            await sleep(0.01)

        match case:
            case "no/sync":
                _ = add_listener(event, listener_sync, ignore=ValueError)
            case "no/async":
                _ = add_listener(event, listener_async, ignore=ValueError)
            case "have/sync":
                _ = add_listener(
                    event, listener_sync, error=error_sync, ignore=ValueError
                )
            case "have/async/sync":
                _ = add_listener(
                    event, listener_async, error=error_sync, ignore=ValueError
                )
            case "have/async":
                _ = add_listener(
                    event, listener_async, error=error_async, ignore=ValueError
                )
        event.emit(True)  # noqa: FBT003
        await sleep(0.01)
        assert called
        assert log == set()
        event.emit(False)  # noqa: FBT003
        await sleep(0.01)
        assert log == set()

    def test_decorators(self) -> None:
        event = Event()
        counter = 0

        def listener() -> None:
            nonlocal counter
            counter += 1

        def increment[**P, R](func: Callable[P, R], /) -> Callable[P, R]:
            @wraps(func)
            def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
                nonlocal counter
                counter += 1
                return func(*args, **kwargs)

            return wrapped

        _ = add_listener(event, listener, decorators=increment)
        event.emit()
        assert counter == 2


class TestLiftListener:
    def test_error(self) -> None:
        def listener() -> None:
            pass

        async def error(event: Event, exception: BaseException, /) -> None:
            _ = (event, exception)
            await sleep(0.01)

        with raises(
            LiftListenerError,
            match="Synchronous listener .* cannot be paired with an asynchronous error handler .*",
        ):
            _ = lift_listener(listener, Event(), error=error, logger_allow_pytest=True)


class TestLiftedEvent:
    def test_main(self) -> None:
        event1 = Event()
        counter = 0

        def listener() -> None:
            nonlocal counter
            counter += 1

        _ = event1.connect(listener)
        event1.emit()
        assert counter == 1

        event2 = Event()

        class Example(LiftedEvent[Callable[[], None]]): ...

        lifted = Example(event=event2)
        _ = lifted.connect(listener)
        lifted.emit()
        assert counter == 2

        def incorrect(x: int, /) -> None:
            assert x >= 0

        _ = lifted.connect(incorrect)  # pyright: ignore[reportArgumentType]


class TestTypedEvent:
    def test_main(self) -> None:
        class Example(TypedEvent[Callable[[int], None]]): ...

        event = Example()

        def correct(x: int, /) -> None:
            assert x >= 0

        _ = event.connect(correct)

        def incorrect(x: int, y: int, /) -> None:
            assert x >= 0
            assert y >= 0

        _ = event.connect(incorrect)  # pyright: ignore[reportArgumentType]



from __future__ import annotations

from utilities.asyncio import sleep_td
from utilities.atools import call_memoized, memoize
from utilities.whenever import SECOND

_DELTA = 0.1 * SECOND


class TestCallMemoized:
    async def test_main(self) -> None:
        counter = 0

        async def increment() -> int:
            nonlocal counter
            counter += 1
            return counter

        for i in range(1, 11):
            assert (await call_memoized(increment)) == i
            assert counter == i

    async def test_refresh(self) -> None:
        counter = 0

        async def increment() -> int:
            nonlocal counter
            counter += 1
            return counter

        for _ in range(2):
            assert (await call_memoized(increment, _DELTA)) == 1
            assert counter == 1
        await sleep_td(2 * _DELTA)
        for _ in range(2):
            assert (await call_memoized(increment, _DELTA)) == 2
            assert counter == 2


class TestMemoize:
    async def test_main(self) -> None:
        counter = 0

        @memoize
        async def increment() -> int:
            nonlocal counter
            counter += 1
            return counter

        for _ in range(10):
            assert await increment() == 1
            assert counter == 1

    async def test_with_arguments(self) -> None:
        counter = 0

        @memoize(duration=_DELTA)
        async def increment() -> int:
            nonlocal counter
            counter += 1
            return counter

        assert await increment() == 1
        assert counter == 1
        await sleep_td(2 * _DELTA)
        assert await increment() == 2
        assert counter == 2



from __future__ import annotations

from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.strategies import binary, booleans

from utilities.gzip import read_binary, write_binary
from utilities.hypothesis import temp_paths

if TYPE_CHECKING:
    from pathlib import Path


class TestReadAndWriteBinary:
    @given(root=temp_paths(), data=binary(), compress=booleans())
    def test_main(self, *, root: Path, data: bytes, compress: bool) -> None:
        file = root.joinpath("file.json")
        write_binary(data, file, compress=compress)
        contents = read_binary(file, decompress=compress)
        assert contents == data



from __future__ import annotations

import asyncio
from asyncio import (
    Lock,
    Queue,
    QueueEmpty,
    Semaphore,
    StreamReader,
    Task,
    TaskGroup,
    create_subprocess_shell,
    sleep,
)
from contextlib import (
    AbstractAsyncContextManager,
    AsyncExitStack,
    _AsyncGeneratorContextManager,
    asynccontextmanager,
    suppress,
)
from dataclasses import dataclass
from io import StringIO
from pathlib import Path
from subprocess import PIPE
from sys import stderr, stdout
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Self,
    TextIO,
    assert_never,
    overload,
    override,
)

from utilities.functions import ensure_int, ensure_not_none
from utilities.os import is_pytest
from utilities.random import SYSTEM_RANDOM
from utilities.sentinel import Sentinel, sentinel
from utilities.shelve import yield_shelf
from utilities.text import to_bool
from utilities.warnings import suppress_warnings
from utilities.whenever import get_now, round_date_or_date_time, to_nanoseconds

if TYPE_CHECKING:
    from asyncio import _CoroutineLike
    from asyncio.subprocess import Process
    from collections.abc import (
        AsyncIterator,
        Callable,
        ItemsView,
        Iterable,
        Iterator,
        KeysView,
        Sequence,
        ValuesView,
    )
    from contextvars import Context
    from random import Random
    from shelve import Shelf
    from types import TracebackType

    from whenever import ZonedDateTime

    from utilities.shelve import _Flag
    from utilities.types import (
        Coro,
        Delta,
        MaybeCallableBoolLike,
        MaybeType,
        PathLike,
        SupportsKeysAndGetItem,
    )


class AsyncDict[K, V]:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, map: SupportsKeysAndGetItem[K, V], /) -> None: ...
    @overload
    def __init__(self, iterable: Iterable[tuple[K, V]], /) -> None: ...
    @override
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__()
        self._dict = dict[K, V](*args, **kwargs)
        self._lock = Lock()

    async def __aenter__(self) -> dict[K, V]:
        await self._lock.__aenter__()
        return self._dict

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
        /,
    ) -> None:
        await self._lock.__aexit__(exc_type, exc, tb)

    def __contains__(self, key: Any, /) -> bool:
        return key in self._dict

    @override
    def __eq__(self, other: Any, /) -> bool:
        return self._dict == other

    __hash__: ClassVar[None] = None  # pyright: ignore[reportIncompatibleMethodOverride]

    def __getitem__(self, key: K, /) -> V:
        return self._dict[key]

    def __iter__(self) -> Iterator[K]:
        yield from self._dict

    def __len__(self) -> int:
        return len(self._dict)

    @override
    def __repr__(self) -> str:
        return repr(self._dict)

    def __reversed__(self) -> Iterator[K]:
        return reversed(self._dict)

    @override
    def __str__(self) -> str:
        return str(self._dict)

    @property
    def empty(self) -> bool:
        return len(self) == 0

    @classmethod
    @overload
    def fromkeys[T](
        cls, iterable: Iterable[T], value: None = None, /
    ) -> AsyncDict[T, Any | None]: ...
    @classmethod
    @overload
    def fromkeys[K2, V2](
        cls, iterable: Iterable[K2], value: V2, /
    ) -> AsyncDict[K2, V2]: ...
    @classmethod
    def fromkeys(
        cls, iterable: Iterable[Any], value: Any = None, /
    ) -> AsyncDict[Any, Any]:
        return cls(dict.fromkeys(iterable, value))

    async def clear(self) -> None:
        async with self._lock:
            self._dict.clear()

    def copy(self) -> Self:
        return type(self)(self._dict.items())

    async def del_(self, key: K, /) -> None:
        async with self._lock:
            del self._dict[key]

    @overload
    def get(self, key: K, default: None = None, /) -> V | None: ...
    @overload
    def get(self, key: K, default: V, /) -> V: ...
    @overload
    def get[V2](self, key: K, default: V2, /) -> V | V2: ...
    def get(self, key: K, default: Any = sentinel, /) -> Any:
        match default:
            case Sentinel():
                return self._dict.get(key)
            case _:
                return self._dict.get(key, default)

    def keys(self) -> KeysView[K]:
        return self._dict.keys()

    def items(self) -> ItemsView[K, V]:
        return self._dict.items()

    @overload
    async def pop(self, key: K, /) -> V: ...
    @overload
    async def pop(self, key: K, default: V, /) -> V: ...
    @overload
    async def pop[V2](self, key: K, default: V2, /) -> V | V2: ...
    async def pop(self, key: K, default: Any = sentinel, /) -> Any:
        async with self._lock:
            match default:
                case Sentinel():
                    return self._dict.pop(key)
                case _:
                    return self._dict.pop(key, default)

    async def popitem(self) -> tuple[K, V]:
        async with self._lock:
            return self._dict.popitem()

    async def set(self, key: K, value: V, /) -> None:
        async with self._lock:
            self._dict[key] = value

    async def setdefault(self, key: K, default: V, /) -> V:
        async with self._lock:
            return self._dict.setdefault(key, default)

    @overload
    async def update(self, m: SupportsKeysAndGetItem[K, V], /) -> None: ...
    @overload
    async def update(self, m: Iterable[tuple[K, V]], /) -> None: ...
    async def update(self, *args: Any, **kwargs: V) -> None:
        async with self._lock:
            self._dict.update(*args, **kwargs)

    def values(self) -> ValuesView[V]:
        return self._dict.values()


##


class EnhancedTaskGroup(TaskGroup):
    """Task group with enhanced features."""

    _max_tasks: int | None
    _semaphore: Semaphore | None
    _timeout: Delta | None
    _error: MaybeType[BaseException]
    _debug: MaybeCallableBoolLike
    _stack: AsyncExitStack
    _timeout_cm: _AsyncGeneratorContextManager[None] | None

    @override
    def __init__(
        self,
        *,
        max_tasks: int | None = None,
        timeout: Delta | None = None,
        error: MaybeType[BaseException] = TimeoutError,
        debug: MaybeCallableBoolLike = False,
    ) -> None:
        super().__init__()
        self._max_tasks = max_tasks
        if (max_tasks is None) or (max_tasks <= 0):
            self._semaphore = None
        else:
            self._semaphore = Semaphore(max_tasks)
        self._timeout = timeout
        self._error = error
        self._debug = debug
        self._stack = AsyncExitStack()
        self._timeout_cm = None

    @override
    async def __aenter__(self) -> Self:
        _ = await self._stack.__aenter__()
        return await super().__aenter__()

    @override
    async def __aexit__(
        self,
        et: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        _ = await self._stack.__aexit__(et, exc, tb)
        match self._is_debug():
            case True:
                with suppress(Exception):
                    _ = await super().__aexit__(et, exc, tb)
            case False:
                _ = await super().__aexit__(et, exc, tb)
            case never:
                assert_never(never)

    @override
    def create_task[T](
        self,
        coro: _CoroutineLike[T],
        *,
        name: str | None = None,
        context: Context | None = None,
    ) -> Task[T]:
        if self._semaphore is None:
            coroutine = coro
        else:
            coroutine = self._wrap_with_semaphore(self._semaphore, coro)
        coroutine = self._wrap_with_timeout(coroutine)
        return super().create_task(coroutine, name=name, context=context)

    def create_task_context[T](self, cm: AbstractAsyncContextManager[T], /) -> Task[T]:
        """Have the TaskGroup start an asynchronous context manager."""
        _ = self._stack.push_async_callback(cm.__aexit__, None, None, None)
        return self.create_task(cm.__aenter__())

    async def run_or_create_many_tasks[**P, T](
        self,
        make_coro: Callable[P, _CoroutineLike[T]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T | Sequence[Task[T]]:
        match self._is_debug(), self._max_tasks:
            case (True, _) | (False, None):
                return await make_coro(*args, **kwargs)
            case False, int():
                return [
                    self.create_task(make_coro(*args, **kwargs))
                    for _ in range(self._max_tasks)
                ]
            case never:
                assert_never(never)

    async def run_or_create_task[T](
        self,
        coro: _CoroutineLike[T],
        *,
        name: str | None = None,
        context: Context | None = None,
    ) -> T | Task[T]:
        match self._is_debug():
            case True:
                return await coro
            case False:
                return self.create_task(coro, name=name, context=context)
            case never:
                assert_never(never)

    def _is_debug(self) -> bool:
        return to_bool(self._debug) or (
            (self._max_tasks is not None) and (self._max_tasks <= 0)
        )

    async def _wrap_with_semaphore[T](
        self, semaphore: Semaphore, coroutine: _CoroutineLike[T], /
    ) -> T:
        async with semaphore:
            return await coroutine

    async def _wrap_with_timeout[T](self, coroutine: _CoroutineLike[T], /) -> T:
        async with timeout_td(self._timeout, error=self._error):
            return await coroutine


##


def get_coroutine_name(func: Callable[[], Coro[Any]], /) -> str:
    """Get the name of a coroutine, and then dispose of it gracefully."""
    coro = func()
    name = coro.__name__
    with suppress_warnings(
        message="coroutine '.*' was never awaited", category=RuntimeWarning
    ):
        del coro
    return name


##


async def get_items[T](queue: Queue[T], /, *, max_size: int | None = None) -> list[T]:
    """Get items from a queue; if empty then wait."""
    try:
        items = [await queue.get()]
    except RuntimeError as error:  # pragma: no cover
        if (not is_pytest()) or (error.args[0] != "Event loop is closed"):
            raise
        return []
    max_size_use = None if max_size is None else (max_size - 1)
    items.extend(get_items_nowait(queue, max_size=max_size_use))
    return items


def get_items_nowait[T](queue: Queue[T], /, *, max_size: int | None = None) -> list[T]:
    """Get items from a queue; no waiting."""
    items: list[T] = []
    if max_size is None:
        while True:
            try:
                items.append(queue.get_nowait())
            except QueueEmpty:
                break
    else:
        while len(items) < max_size:
            try:
                items.append(queue.get_nowait())
            except QueueEmpty:
                break
    return items


##


async def put_items[T](items: Iterable[T], queue: Queue[T], /) -> None:
    """Put items into a queue; if full then wait."""
    for item in items:
        await queue.put(item)


def put_items_nowait[T](items: Iterable[T], queue: Queue[T], /) -> None:
    """Put items into a queue; no waiting."""
    for item in items:
        queue.put_nowait(item)


##


async def sleep_max(
    sleep: Delta | None = None, /, *, random: Random = SYSTEM_RANDOM
) -> None:
    """Sleep which accepts deltas."""
    if sleep is None:
        return
    await asyncio.sleep(random.uniform(0.0, to_nanoseconds(sleep) / 1e9))


##


async def sleep_rounded(delta: Delta, /) -> None:
    """Sleep until a rounded time."""
    await sleep_until(round_date_or_date_time(get_now(), delta, mode="ceil"))


##


async def sleep_td(delta: Delta | None = None, /) -> None:
    """Sleep which accepts deltas."""
    if delta is None:
        return
    await sleep(to_nanoseconds(delta) / 1e9)


##


async def sleep_until(datetime: ZonedDateTime, /) -> None:
    """Sleep until a given time."""
    await sleep_td(datetime - get_now())


##


@dataclass(kw_only=True, slots=True)
class StreamCommandOutput:
    process: Process
    stdout: str
    stderr: str

    @property
    def return_code(self) -> int:
        return ensure_int(self.process.returncode)  # skipif-not-windows


async def stream_command(cmd: str, /) -> StreamCommandOutput:
    """Run a shell command asynchronously and stream its output in real time."""
    process = await create_subprocess_shell(  # skipif-not-windows
        cmd, stdout=PIPE, stderr=PIPE
    )
    proc_stdout = ensure_not_none(  # skipif-not-windows
        process.stdout, desc="process.stdout"
    )
    proc_stderr = ensure_not_none(  # skipif-not-windows
        process.stderr, desc="process.stderr"
    )
    ret_stdout = StringIO()  # skipif-not-windows
    ret_stderr = StringIO()  # skipif-not-windows
    async with TaskGroup() as tg:  # skipif-not-windows
        _ = tg.create_task(_stream_one(proc_stdout, stdout, ret_stdout))
        _ = tg.create_task(_stream_one(proc_stderr, stderr, ret_stderr))
    _ = await process.wait()  # skipif-not-windows
    return StreamCommandOutput(  # skipif-not-windows
        process=process, stdout=ret_stdout.getvalue(), stderr=ret_stderr.getvalue()
    )


async def _stream_one(
    input_: StreamReader, out_stream: TextIO, ret_stream: StringIO, /
) -> None:
    """Asynchronously read from a stream and write to the target output stream."""
    while True:  # skipif-not-windows
        line = await input_.readline()
        if not line:
            break
        decoded = line.decode()
        _ = out_stream.write(decoded)
        out_stream.flush()
        _ = ret_stream.write(decoded)


##


@asynccontextmanager
async def timeout_td(
    timeout: Delta | None = None, /, *, error: MaybeType[BaseException] = TimeoutError
) -> AsyncIterator[None]:
    """Timeout context manager which accepts deltas."""
    timeout_use = None if timeout is None else (to_nanoseconds(timeout) / 1e9)
    try:
        async with asyncio.timeout(timeout_use):
            yield
    except TimeoutError:
        raise error from None


##


_LOCKS: AsyncDict[Path, Lock] = AsyncDict()


@asynccontextmanager
async def yield_locked_shelf(
    path: PathLike,
    /,
    *,
    flag: _Flag = "c",
    protocol: int | None = None,
    writeback: bool = False,
) -> AsyncIterator[Shelf[Any]]:
    """Yield a shelf, behind a lock."""
    path = Path(path)
    try:
        lock = _LOCKS[path]
    except KeyError:
        lock = Lock()
        await _LOCKS.set(path, lock)
    async with lock:
        with yield_shelf(
            path, flag=flag, protocol=protocol, writeback=writeback
        ) as shelf:
            yield shelf


__all__ = [
    "AsyncDict",
    "EnhancedTaskGroup",
    "StreamCommandOutput",
    "get_coroutine_name",
    "get_items",
    "get_items_nowait",
    "put_items",
    "put_items_nowait",
    "sleep_max",
    "sleep_rounded",
    "sleep_td",
    "sleep_until",
    "stream_command",
    "timeout_td",
    "yield_locked_shelf",
]



from __future__ import annotations

from random import Random, SystemRandom
from re import search
from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.strategies import integers, iterables, just, sampled_from

from utilities.random import (
    SYSTEM_RANDOM,
    bernoulli,
    get_docker_name,
    get_state,
    shuffle,
)

if TYPE_CHECKING:
    from collections.abc import Iterable


class TestBernoulli:
    @given(case=sampled_from([(0.0, False), (1.0, True)]))
    def test_main(self, *, case: tuple[float, bool]) -> None:
        true, expected = case
        result = bernoulli(true=true)
        assert result is expected


class TestGetDockerName:
    def test_main(self) -> None:
        name = get_docker_name()
        assert search(r"^[a-z]+_[a-z]+\d$", name)

    def test_deterministic(self) -> None:
        expected = [
            "priceless_hypatia8",
            "quizzical_aryabhata6",
            "priceless_wing2",
            "dazzling_mclean2",
            "epic_montalcini8",
        ]
        for exp in expected:
            result = get_docker_name(TestGetDockerName.test_deterministic.__qualname__)
            assert result == exp


class TestGetState:
    @given(seed=integers() | just(SYSTEM_RANDOM))
    def test_main(self, *, seed: int | SystemRandom) -> None:
        state = get_state(seed)
        assert isinstance(state, Random)

    def test_deterministic(self) -> None:
        expected = [6, 4, 8, 10, 6]
        for exp in expected:
            result = get_state(TestGetState.test_deterministic.__qualname__).randint(
                0, 10
            )
            assert result == exp


class TestShuffle:
    @given(iterable=iterables(integers()), seed=integers())
    def test_main(self, *, iterable: Iterable[int], seed: int) -> None:
        as_set = set(iterable)
        result = shuffle(as_set, seed=seed)
        assert set(result) == as_set

    def test_deterministic(self) -> None:
        expected = [
            [2, 3, 0, 4, 1],
            [3, 1, 2, 0, 4],
            [1, 0, 4, 3, 2],
            [1, 0, 2, 3, 4],
            [4, 0, 3, 1, 2],
        ]
        for exp in expected:
            result = shuffle(range(5), seed=TestShuffle.test_deterministic.__qualname__)
            assert result == exp


class TestSystemRandom:
    def test_main(self) -> None:
        assert isinstance(SYSTEM_RANDOM, SystemRandom)



from __future__ import annotations

from os import getenv

from hypothesis import given
from hypothesis.strategies import (
    DataObject,
    booleans,
    data,
    integers,
    none,
    sampled_from,
)
from pytest import raises

from utilities.hypothesis import text_ascii
from utilities.os import (
    CPU_COUNT,
    GetCPUUseError,
    GetEnvVarError,
    get_cpu_count,
    get_cpu_use,
    get_env_var,
    is_debug,
    is_pytest,
    temp_environ,
)
from utilities.pytest import skipif_windows

text = text_ascii(min_size=1, max_size=10)


def _prefix(text: str, /) -> str:
    return f"_TEST_OS_{text}"


class TestGetCPUCount:
    def test_function(self) -> None:
        assert isinstance(get_cpu_count(), int)

    def test_constant(self) -> None:
        assert isinstance(CPU_COUNT, int)


class TestGetCPUUse:
    @given(n=integers(min_value=1))
    def test_int(self, *, n: int) -> None:
        result = get_cpu_use(n=n)
        assert result == n

    def test_all(self) -> None:
        result = get_cpu_use(n="all")
        assert isinstance(result, int)
        assert result >= 1

    @given(n=integers(max_value=0))
    def test_error(self, *, n: int) -> None:
        with raises(GetCPUUseError, match=r"Invalid number of CPUs to use: -?\d+"):
            _ = get_cpu_use(n=n)


class TestGetEnvVar:
    @given(
        key=text.map(_prefix), value=text, default=text | none(), nullable=booleans()
    )
    @skipif_windows
    def test_case_sensitive(
        self, *, key: str, value: str, default: str | None, nullable: bool
    ) -> None:
        with temp_environ({key: value}):
            result = get_env_var(key, default=default, nullable=nullable)
        assert result == value

    @given(
        data=data(),
        key=text.map(_prefix),
        value=text,
        default=text | none(),
        nullable=booleans(),
    )
    def test_case_insensitive(
        self,
        *,
        data: DataObject,
        key: str,
        value: str,
        default: str | None,
        nullable: bool,
    ) -> None:
        key_use = data.draw(sampled_from([key, key.lower(), key.upper()]))
        with temp_environ({key: value}):
            result = get_env_var(key_use, default=default, nullable=nullable)
        assert result == value

    @given(
        key=text.map(_prefix),
        case_sensitive=booleans(),
        default=text,
        nullable=booleans(),
    )
    def test_default(
        self, *, key: str, case_sensitive: bool, default: str, nullable: bool
    ) -> None:
        value = get_env_var(
            key, case_sensitive=case_sensitive, default=default, nullable=nullable
        )
        assert value == default

    @given(key=text.map(_prefix), case_sensitive=booleans())
    def test_nullable(self, *, key: str, case_sensitive: bool) -> None:
        value = get_env_var(key, case_sensitive=case_sensitive, nullable=True)
        assert value is None

    @given(key=text.map(_prefix), case_sensitive=booleans())
    def test_error(self, *, key: str, case_sensitive: bool) -> None:
        with raises(
            GetEnvVarError, match=r"No environment variable .*(\(modulo case\))?"
        ):
            _ = get_env_var(key, case_sensitive=case_sensitive)


class TestIsDebug:
    def test_main(self) -> None:
        result = is_debug()
        assert isinstance(result, bool)


class TestIsPytest:
    def test_main(self) -> None:
        assert is_pytest()


class TestTempEnviron:
    @given(key=text.map(_prefix), value=text)
    def test_set(self, *, key: str, value: str) -> None:
        assert getenv(key) is None
        with temp_environ({key: value}):
            assert getenv(key) == value
        assert getenv(key) is None

    @given(key=text.map(_prefix), prev=text, new=text)
    def test_override(self, *, key: str, prev: str, new: str) -> None:
        with temp_environ({key: prev}):
            assert getenv(key) == prev
            with temp_environ({key: new}):
                assert getenv(key) == new
            assert getenv(key) == prev

    @given(key=text.map(_prefix), value=text)
    def test_unset(self, *, key: str, value: str) -> None:
        with temp_environ({key: value}):
            assert getenv(key) == value
            with temp_environ({key: None}):
                assert getenv(key) is None
            assert getenv(key) == value



from __future__ import annotations

from ipaddress import IPv4Address

from utilities.http import get_public_ip, yield_connection
from utilities.pytest import throttle
from utilities.whenever import MINUTE


class TestGetPublicIP:
    @throttle(delta=5 * MINUTE)
    def test_main(self) -> None:
        ip = get_public_ip(timeout=10.0)
        assert isinstance(ip, IPv4Address)


class TestYieldConnection:
    def test_main(self) -> None:
        with yield_connection("api.ipify.org"):
            pass
